from ._ge import _GeoEnrichment
from arcgis import __version__
from arcgis import env
from arcgis.geometry import Geometry
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._deprecate import deprecated
from arcgis._impl.common._utils import _lazy_property
import re

import collections

BufferStudyArea = collections.namedtuple('BufferStudyArea', 'area radii units overlap travel_mode')
BufferStudyArea.__new__.__defaults__ = (None, None, None, True, None)
BufferStudyArea.__doc__ = """BufferStudyArea allows you to buffer point and street address study areas.

Parameters:
area: the point geometry or street address (string) study area to be buffered
radii: list of distances by which to buffer the study area, eg. [1, 2, 3]
units: distance unit, eg. Miles, Kilometers, Minutes (when using drive times/travel_mode)
overlap: boolean, uses overlapping rings when True, or non-overlapping disks when False
travel_mode: None or string, one of the supported travel modes when using network service areas, eg. Driving, Trucking, Walking.
"""


def _pep8ify(name):
    """PEP8ify name"""
    if '.' in name:
        name = name[name.rfind('.') + 1:]
    if name[0].isdigit():
        name = "level_" + name
    name = name.replace(".", "_")
    if '_' in name:
        return name.lower()
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class NamedArea(object):
    """
    Represents named geographical places in a country. Each named area has attributes for the
    supported subgeography levels within it, and the value of those attributes are dictionaries containing the named
    places within that level of geography. This allows for interactive selection of places using intellisense and a
    notation such as the following:
    .. code-block:: python

        # Usage Example

        usa = Country.get('USA')
        usa.subgeographies.states['California'].counties['San_Bernardino_County']

    """
    def __init__(self, country, name=None, level=None, areaid='01', geometry=None):
        self._gis = country._gis
        self._country = country
        self._currlvl = level
        self._areaid = areaid
        if geometry is not None:
            self.geometry = geometry
        if name is None:
            name = country.properties.name
        self._name = name
        self._level_mappings = {}

        for childlevel in self._childlevels:
            setattr(self, childlevel, None)

    @property
    def __studyarea__(self):
        return {"sourceCountry": self._country.properties.id, "layer": self._currlvl,"ids":[self._areaid]}

    def __str__(self):
        return '<%s name:"%s" area_id="%s", level="%s", country="%s">' % (type(self).__name__, self._name, self._areaid,\
                                                                         self._currlvl, self._country.properties.name)
    def __repr__(self):
        return '<%s name:"%s" area_id="%s", level="%s", country="%s">' % (type(self).__name__, self._name, self._areaid,\
                                                                         self._currlvl, self._country.properties.name)

    @property
    def _childlevels(self):
        dset = [dset for dset in self._country._geog_levels if dset['datasetID'] == self._country.dataset][0]
        whole_country_levelid = [lvl['id'] for lvl in dset['levels'] if lvl['isWholeCountry']][0]
        if self._currlvl is None:
            self._currlvl = whole_country_levelid

        is_whole_country = self._currlvl == whole_country_levelid

        childlevels = set()
        for branch in dset['branches']:

            levels = branch['levels']
            if is_whole_country and self._currlvl not in levels:
                level_attr = _pep8ify(levels[0])
                childlevels.add(level_attr)
                self._level_mappings[level_attr] = levels[0]

            elif self._currlvl in levels:
                try:
                    nextlevel = levels[levels.index(self._currlvl) + 1]
                    level_attr = _pep8ify(nextlevel)
                    childlevels.add(level_attr)
                    self._level_mappings[level_attr] = nextlevel
                except IndexError:
                    # no nextlevel
                    pass
        return childlevels

    def __getattribute__(self, name):
        if not name.startswith('_') and not name in ['geometry']:
            val = object.__getattribute__(self, name)
            if val is None:
                # print('Fetching {}'.format(name))
                self._fetch_subgeographies(name)
            return object.__getattribute__(self, name)

        else:
            return object.__getattribute__(self, name)


    def _fetch_subgeographies(self, name):
        df = standard_geography_query(source_country=self._country.properties.id,
                                      layers=[self._currlvl],
                                      ids=[self._areaid],
                                      return_sub_geography=True,
                                      sub_geography_layer=self._level_mappings[name],
                                      return_geometry=True,
                                      as_featureset=False)

        places = {}
        for index, row in df.iterrows():
            #     print(dict(row))
            plc = dict(row)
            place = NamedArea(country=self._country, name=plc['AreaName'], level=plc['DataLayerID'],
                              areaid=plc['AreaID'], geometry=Geometry(plc['SHAPE']))
            place_name = plc['AreaName'].replace(' ', '_')
            if self._level_mappings[name] == 'US.ZIP5':
                place_name = plc['AreaID']
            places[place_name] = place
        setattr(self, name, places)
#----------------------------------------------------------------------
class Country(object):
    """
    A country for which geoenrichment data is available. The Country class can be used
    to discover the data collections, sub-geographies and available reports for a country.
    """
    @classmethod
    def get(cls, name):
        """
        Gets a reference to a particular country, given its name, or its
        2 letter abbreviation or ISO3 code.

        ================  ========================================================
        **Argument**      **Description**
        ----------------  --------------------------------------------------------
        name              Required string. The country name or 2 letter/ISO3 code
        ================  ========================================================

        Returns the country
        """
        cs = get_countries()
        if len(name) == 2:
            return [c for c in cs if c.properties.id == name][0]
        elif len(name) == 3:
            return [c for c in cs if c.properties.abbr3 == name][0]
        else:
            cnames = [c for c in cs if c.properties.name.upper() == name.upper()]
            if len(cnames) == 1:
                return cnames[0]
            else:
                altnames = [c for c in cs if c.properties.altName.upper() == name.upper()]
                if len(altnames) == 1:
                    return altnames[0]
                else:
                    raise ValueError('Unable to find country with the specified name, id, ISO 3 country code')

    # noinspection PyMissingConstructor
    def __init__(self, dictdata, gis, purl=None):
        self._gis = gis
        if self._gis._is_hosted_nb_home == False:
            hs = dict(self._gis.properties['helperServices'])
            if 'geoenrichment' in hs:
                self._base_url = hs['geoenrichment']['url']
            else:
                self._base_url = 'http://geoenrich.arcgis.com/arcgis/rest/services/World/geoenrichmentserver'
        elif self._gis._is_hosted_nb_home and purl:
            self._base_url = purl
        else:
            hs = dict(self._gis.properties['helperServices'])
            if 'geoenrichment' in hs:
                self._base_url = hs['geoenrichment']['url']
            else:
                self._base_url = 'http://geoenrich.arcgis.com/arcgis/rest/services/World/geoenrichmentserver'
            if self._gis._is_hosted_nb_home:
                self._base_url = self._validate_url(self._base_url)
        self.properties = PropertyMap(dictdata)
        self._dataset_id = self.properties.defaultDatasetID
    #----------------------------------------------------------------------
    def _validate_url(self, url):
        res = self._gis._private_service_url(url)
        if "privateServiceUrl" in res:
            return res["privateServiceUrl"]
        else:
            return res["serviceUrl"]
        return url
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s name:%s>' % (type(self).__name__, self.properties.name)

    def __repr__(self):
        return '<%s name:%s>' % (type(self).__name__, self.properties.name)

    @_lazy_property
    def geometry(self):
        lvlid = [lvl['id'] for lvl in self.levels if lvl['isWholeCountry']][0]
        df= standard_geography_query(source_country=self.properties.id,
                                        layers=[lvlid],
                                        ids=['01'],
                                        return_sub_geography=False,
                                        return_geometry=True,
                                        as_featureset=False)
        return Geometry(df.iloc[0]['SHAPE'])



    @_lazy_property
    def _geog_levels(self):
        """
        Returns levels of geography in this country, including branches for all datasets
        """
        params = {'f': 'json'}
        url = self._base_url + "/Geoenrichment/standardgeographylevels/%s" % (self.properties.id)
        res = self._gis._con.post(url, params)
        return res['geographyLevels'][0]['datasets']

    @property
    def levels(self):
        """
        Returns levels of geography in this country, for the current dataset
        """
        dset = [d for d in self._geog_levels if d['datasetID'] == self._dataset_id][0]
        return dset['levels']

    @property
    def dataset(self):
        """
        Returns the currently used dataset for this country
        """
        return self._dataset_id

    @dataset.setter
    def dataset(self, value):
        if value in self.properties.datasets:
            self._dataset_id = value
            try:
                delattr(self, '_lazy_subgeographies')
                delattr(self, '_lazy__geog_levels')
            except:
                pass
        else:
            raise ValueError('The specified dataset is not available in this country. Choose one of '+ str(self.properties.datasets))

    @_lazy_property
    def data_collections(self):
        """
        Returns the supported data collections and analysis variables as a Pandas dataframe.

        The dataframe is indexed by the data collection id(dataCollectionID) and contains columns for
        analysis variables(analysisVariable)
        """
        import pandas as pd
        df = pd.io.json.json_normalize((_data_collections(country=self.properties.id,
                out_fields=['id', 'dataCollectionID', 'alias', 'fieldCategory', 'vintage']))['DataCollections'], 'data', 'dataCollectionID')
        df['analysisVariable'] = df['dataCollectionID'] + '.' + df['id']
        df = df[['dataCollectionID', 'analysisVariable', 'alias', 'fieldCategory', 'vintage']]
        df.set_index('dataCollectionID', inplace=True)
        return df

    @_lazy_property
    def subgeographies(self):
        """
        Returns the named geographical places in this country, as NamedArea objects. Each named area has attributes for the
        supported subgeography levels within it, and the value of those attributes are dictionaries containing the named
        places within that level of geography. This allows for interactive selection of places using intellisense and a
        notation such as the following:

        .. code-block:: python

            # Usage Example 1

            usa = Country.get('USA')
            usa.subgeographies.states['California'].counties['San_Bernardino_County']

    .. code-block:: python

            # Usage Example 2

            india.named_places.states['Bihar'].districts['Aurangabad'].subdistricts['Barun']

        """
        return NamedArea(self)

    def search(self, query, layers=['*']):
        """
        Searches this country for places that have the specified query string in their name.

        Returns a list of named areas matching the specified query

        ================  ========================================================
        **Argument**      **Description**
        ----------------  --------------------------------------------------------
        query             Required string. The query string to search for places
                          within this country.
        ----------------  --------------------------------------------------------
        levels            Optional list of layer ids. Layer ids for a country
                          can be queried using Country.levels properties.
        ================  ========================================================

        :return:
            A list of named areas that match the query string
        """
        df = standard_geography_query(source_country=self.properties.id, geoquery=query,
                                      layers=layers,
                                      return_geometry=True,
                                      as_featureset=False)

        places = []
        for index, row in df.iterrows():
            plc = dict(row)
            place = NamedArea(country=self, name=plc['AreaName'], level=plc['DataLayerID'],
                              areaid=plc['AreaID'], geometry=plc['SHAPE'])
            places.append(place)

        return places

    @_lazy_property
    def reports(self):
        """Returns the available reports for this country as a Pandas dataframe"""
        import pandas as pd
        rdf = _find_report(self.properties.id)
        df = pd.io.json.json_normalize(rdf)
        df = df[['reportID', 'metadata.title', 'metadata.categories', 'formats']].rename(
            columns={'reportID': 'id', 'metadata.title': 'title', 'metadata.categories': 'categories'})
        return df



def get_countries(gis=None):
    """Returns the countries for which there is GeoEnrichment data."""
    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)
    if gis._is_hosted_nb_home:
        return [Country(c, gis, purl=ge._base_url) for c in ge.countries(as_df=False)]
    else:
        return [Country(c, gis, purl=None) for c in ge.countries(as_df=False)]

def create_report(study_areas,
                  report=None,
                  export_format='pdf',
                  report_fields=None,
                  options=None,
                  return_type=None,
                  use_data=None,
                  in_sr=4326,
                  out_name=None,
                  out_folder=None,
                  gis=None):
    """
    The Create Report method allows you to create many types of high quality reports for a
    variety of use cases describing the input area. If a point is used as a study area, the
    service will create a 1-mile ring buffer around the point to collect and append enrichment
    data. Optionally, you can create a buffer ring or drive-time service area around points of
    interest to generate PDF or Excel reports containing relevant information for the area on
    demographics, consumer spending, tapestry market, business or market potential.

    Report options are available and can be used to describe and gain a better understanding
    about the market, customers / clients and competition associated with an area of interest.


    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    study_areas            required list. Required parameter: Study areas may be defined by
                           input points, polygons, administrative boundaries or addresses.
    ------------------     --------------------------------------------------------------------
    report                 optional string. identify the id of the report. This may be one of
                           the many default reports available along with our demographic data
                           collections or a customized report. Custom report templates are
                           stored in an ArcGIS Online organization as a Report Template item.
                           The organization URL and a valid ArcGIS Online authentication token
                           is required for security purposes to access these templates. If no
                           report is specified, the default report is census profile for United
                           States and a general demographic summary report for most countries.
    ------------------     --------------------------------------------------------------------
    export_format          Optional parameter to specify the format of the generated report.
                           Supported formats include PDF and XLSX.
    ------------------     --------------------------------------------------------------------
    report_fields          Optional parameter specifies additional choices to customize
                           reports. Below is an example of the position on the report header
                           for each field.
    ------------------     --------------------------------------------------------------------
    options                Optional parameter to specify the properties for the study area
                           buffer. For a full list of valid buffer properties values and
                           further examples review the Input XY Locations' options parameter.

                           By default a 1 mile radius buffer will be applied to point(s) and
                           address locations to define a study area.
    ------------------     --------------------------------------------------------------------
    return_type            Optional parameter used for storing an output report item to Portal
                           for ArcGIS instead of returning a report to a customer via binary
                           stream. The attributes are used by Portal to determine where and how
                           an item is stored. Parameter attributes include: user, folder,
                           title, item_properties, URL, token, and referrer.
                           Example

                           Creating a new output in a Portal for ArcGIS Instance:

                           return_type = {'user' : 'testUser',
                                          'folder' : 'FolderName',
                                          'title' : 'Report Title',
                                          'item_properties' : '<properties>',
                                          'url' : 'https://hostname.domain.com/webadaptor',
                                          'token' : 'token', 'referrer' : 'referrer'}
    ------------------     --------------------------------------------------------------------
    use_data               Optional dictionary. This parameter explicitly specify the country
                           or dataset to query. When all input features specified in the
                           study_areas parameter describe locations or areas that lie in the
                           same country or dataset, this parameter can be specified to provide
                           an additional 'performance hint' to the service.

                           By default, the service will automatically determine the country or
                           dataset that is associated with each location or area submitted in
                           the study_areas parameter. Specifying a specific dataset or country
                           through this parameter will potentially improve response time.

                           By default, the data apportionment method is determined by the size
                           of the study area. Small study areas use block apportionment for
                           higher accuracy whereas large study areas (100 miles or more) will
                           use a cascading centroid apportionment method to maintain
                           performance. This default behavior can be overridden by using the
                           detailed_aggregation parameter.
    ------------------     --------------------------------------------------------------------
    in_sr                  Optional parameter to define the input geometries in the study_areas
                           parameter in a specified spatial reference system.
                           When input points are defined in the study_areas parameter, this
                           optional parameter can be specified to explicitly indicate the
                           spatial reference system of the point features. The parameter value
                           can be specified as the well-known ID describing the projected
                           coordinate system or geographic coordinate system.
                           The default is 4326
    ------------------     --------------------------------------------------------------------
    out_name               Optional string.  Name of the output file [ending in .pdf or .xlsx)
    ------------------     --------------------------------------------------------------------
    out_folder             Optional string. Name of the save folder
    ==================     ====================================================================
    """
    if gis is None:
        gis = env.active_gis

    areas = []
    for area in study_areas:
        area_dict = area
        if isinstance(area, str):  # street address - {"address":{"text":"380 New York St Redlands CA 92373"}}
            area_dict = {'address': {'text': area}}
        elif isinstance(area, dict):  # pass through - user knows what they're sending
            pass
        elif isinstance(area, Geometry):  # geometry, polygons, points
            area_dict = {'geometry': dict(area)}
        elif isinstance(area, BufferStudyArea):

            # namedtuple('BufferStudyArea', 'area radii units overlap travel_mode')
            g = area.area
            if isinstance(g, str):
                area_dict = {'address': {'text': g}}
            elif isinstance(g, dict):
                area_dict = g
            elif isinstance(g, Geometry):  # geometry, polygons, points
                area_dict = {'geometry': dict(g)}
            else:
                raise ValueError('BufferStudyArea is only supported for Point geometry and addresses')

            area_type = "RingBuffer"
            if area.travel_mode is None:
                if not area.overlap:
                    area_type = "RingBufferBands"
            else:
                area_type = "NetworkServiceArea"

            area_dict['areaType'] = area_type
            area_dict['bufferUnits'] = area.units
            area_dict['bufferRadii'] = area.radii
            if area.travel_mode is not None:
                area_dict['travel_mode'] = area.travel_mode

        elif isinstance(area, NamedArea):  # named area
            area_dict = area.__studyarea__
        elif isinstance(area, list):  # list of named areas, (union)
            first_area = area[0]
            ids = []
            if isinstance(first_area, NamedArea):
                for namedarea in area:
                    a = namedarea.__studyarea__
                    if a['layer'] != first_area['layer'] or a['sourceCountry'] != first_area['sourceCountry']:
                        raise ValueError('All NamedAreas in the list must have the same source country and level')
                    ids.append(a['ids'])
                area_dict = {"sourceCountry": first_area['sourceCountry'], "layer": first_area['layer'],
                             "ids": [ids.join(",")]}
            else:
                raise ValueError('Lists members must be NamedArea instances')
        else:
            raise ValueError("Don't know how to handle study areas of type " + str(type(area)))

        areas.append(area_dict)
    ge = _GeoEnrichment(gis=gis)
    return ge.create_report(study_areas=areas,
                             report=report,
                             export_format=export_format,
                            report_fields=report_fields,
                            options=options,
                            return_type=return_type,
                            use_data=use_data,
                            in_sr=in_sr,
                            out_folder=out_folder,
                            out_name=out_name)
#----------------------------------------------------------------------
def _data_collections(country=None,
                     collection_name=None,
                     variables=None,
                     out_fields="*",
                     hide_nulls=True,
                     gis=None,
                     as_dict=True):
    """
    The GeoEnrichment class uses the concept of a data collection to define the data
    attributes returned by the enrichment service. Each data collection has a unique name
    that acts as an ID that is passed in the data_collections parameter of the GeoEnrichment
    service.

    Some data collections (such as default) can be used in all supported countries. Other data
    collections may only be available in one or a collection of countries. Data collections may
    only be available in a subset of countries because of differences in the demographic data
    that is available for each country. A list of data collections for all available countries
    can be generated with the data collection discover method seen below.
    Return a list of data collections that can be run for any country.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    country                optional string. lets the user supply and optional name of a country
                           in order to get information about the data collections in that given
                           country.
    ------------------     --------------------------------------------------------------------
    dataset                Optional string. Name of the data collection to examine.
    ------------------     --------------------------------------------------------------------
    variables              Optional string/list. This parameter to specifies a list of field
                           names that include variables for the derivative statistics.
    ------------------     --------------------------------------------------------------------
    out_fields             Optional string. This parameter is a string of comma seperate field
                           names.
    ------------------     --------------------------------------------------------------------
    hide_nulls             Optional boolean. parameter to return only values that are not NULL
                           in the output response. Adding the optional suppress_nulls parameter
                           to any data collections discovery method will reduce the size of the
                           output that is returned.
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS.  If None, the GIS object will be used from the
                           arcgis.env.active_gis.  This GIS object must be authenticated and
                           have the ability to consume credits
    ------------------     --------------------------------------------------------------------
    as_dict                Opional boolean. If True, the result comes back as a python
                           dictionary, else the value will returns as a Python DataFrame.
    ==================     ====================================================================

    :returns: dictionary, describing the requested return data.
    """
    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)

    return ge.data_collections(country=country,
                                collection_name=collection_name,
                                variables=variables,
                                out_fields=out_fields,
                                hide_nulls=hide_nulls,
                                as_dict=as_dict)
#----------------------------------------------------------------------
def enrich(study_areas,
           data_collections=None,
           analysis_variables=None,
           comparison_levels=None,
           add_derivative_variables=None,
           intersecting_geographies=None,
           return_geometry=True,
           # options=None, # can be specified in study_areas
           # use_data=None, # is only a 'performance hint'
           # in_sr=4326, # will use the sr from the geometry
           # out_sr=4326, # will use arcgis.env.out_sr
           # suppress_nulls=False, # never
           # for_storage=True, # undocumented, not required
           # as_featureset=True, # always return df
           gis=None):
    """
    Returns demographic and other requested information for the specified study areas.
    Study areas define the location of the point or area that you want to enrich
    with additional information or creare reports about. If one or many points are input as
    a study area, the method will create a 1-mile ring buffer around
    the point to collect and append enrichment data. You can optionally
    change the ring buffer size or create drive-time service areas
    around the point.

    You can create a buffer ring or drive-time service
    area around the points to aggregate data for the study areas. You
    can also return enrichment data for buffers around input line
    features.

    =========================     ====================================================================
    **Argument**                  **Description**
    -------------------------     --------------------------------------------------------------------
    study_areas                   Required list, FeatureSet or SpatiallyEnabledDataFrame containing
                                  the input areas to be enriched.

                                  study_areas can be a SpatiallyEnabledDataFrame, FeatureSet or a
                                  lists of the following types:
                                  * addresses, points of interest, place names or other
                                  supported locations as strings.
                                  * dicts such as [{"address":{"Address":"380 New York St.",
                                  "Admin1":"Redlands","Admin2":"CA","Postal":"92373",
                                  "CountryCode":"USA"}}] for multiple field addresses
                                  * arcgis.gis.Geometry instances
                                  * BufferStudyArea instances. By default, one-mile ring
                                  buffers are created around the points to collect and append
                                  enrichment data. You can use BufferStudyArea to change the ring
                                  buffer size or create drive-time service areas around the points.
                                  * NamedArea instances to support standard geography. They are
                                  obtained using Country.subgeographies()/search(). When
                                  the NamedArea instances should be combined together (union), a list
                                  of such NamedArea instances should constitute a study area in the
                                  list of requested study areas.
    -------------------------     --------------------------------------------------------------------
    data_collections              Optional list. A Data Collection is a preassembled list of
                                  attributes that will be used to enrich the input features.
                                  Enrichment attributes can describe various types of information such
                                  as demographic characteristics and geographic context of the
                                  locations or areas submitted as input features in study_areas.
    -------------------------     --------------------------------------------------------------------
    analysis_variables            Optional list. A Data Collection is a preassembled list of
                                  attributes that will be used to enrich the input features. With the
                                  analysis_variables parameter you can return a subset of variables
                                  enrichment attributes can describe various types of information such
                                  as demographic characteristics and geographic context of the
                                  locations or areas submitted as input features in study_areas.
    -------------------------     --------------------------------------------------------------------
    add_derivative_variables      Optional list. This parameter is used to specify an array of string
                                  values that describe what derivative variables to include in the
                                  output. The list of accepted values includes:
                                  ['percent','index','average','all','*']
    -------------------------     --------------------------------------------------------------------
    comparison_levels             Optional list of layer IDs for which the intersecting
                                  geographies should be geoenriched.
    -------------------------     --------------------------------------------------------------------
    intersecting_geographies      Optional parameter to explicitly define the geographic layers used
                                  to provide geographic context during the enrichment process. For
                                  example, you can use this optional parameter to return the U.S.
                                  county and ZIP Code that each input study area intersects.
                                  You can intersect input features defined in the study_areas
                                  parameter with standard geography layers that are provided by the
                                  GeoEnrichment class for each country. You can also intersect
                                  features from a publicly available feature service.
    -------------------------     --------------------------------------------------------------------
    return_geometry               Optional boolean. A parameter to request the output geometries in
                                  the response.
    -------------------------     --------------------------------------------------------------------
    gis                           Optional GIS.  If None, the GIS object will be used from the
                                  arcgis.env.active_gis.  This GIS object must be authenticated and
                                  have the ability to consume credits
    =========================     ====================================================================

    Refer to https://developers.arcgis.com/rest/geoenrichment/api-reference/street-address-locations.htm for
    the format of intersection_geographies parameter.

    Performance Tip: If you wish to speed up the operation and don't care about the geometries, set
    return_geometry=False

    :returns: Spatial DataFrame or Panda's DataFrame with the requested information for the study areas
    """
    import pandas as pd
    from arcgis.features import SpatialDataFrame, FeatureSet, GeoAccessor, GeoSeriesAccessor

    def _chunks(l, n):
        """yield successive n-sized chunks from l."""
        for i in range(0, len(l), n):
            yield l[i:i + n]

    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)

    areas = study_areas
    if isinstance(study_areas, FeatureSet):
        areas = FeatureSet.sdf
    elif isinstance(study_areas, dict): # could be dict of NamedAreas, eg usa.subgeographies.states['California'].counties
        areas = list(study_areas.values())
        study_areas = areas

    # convert to dict
    # add comparison levels if any
    # add buffer info - network, ring
    if isinstance(areas, list):
        areas = []
        for area in study_areas:
            area_dict = area
            if isinstance(area, str): # street address - {"address":{"text":"380 New York St Redlands CA 92373"}}
                area_dict = {'address': {'text': area}}
            elif isinstance(area, Geometry): # geometry, polygons, points
                area_dict = {'geometry': dict(area)}
            elif isinstance(area, BufferStudyArea):
                # namedtuple('BufferStudyArea', 'area radii units overlap travel_mode')
                g = area.area
                if isinstance(g, str):
                    area_dict = {'address': {'text': g}}
                elif isinstance(g, Geometry):  # geometry, polygons, points
                    area_dict = {'geometry': dict(g)}
                elif isinstance(g, dict):
                    area_dict = g
                else:
                    raise ValueError('BufferStudyArea is only supported for Point geometry and addresses')

                area_type = "RingBuffer"
                if area.travel_mode is None:
                    if not area.overlap:
                        area_type = "RingBufferBands"
                else:
                    area_type = "NetworkServiceArea"

                area_dict['areaType'] = area_type
                area_dict['bufferUnits'] = area.units
                area_dict['bufferRadii'] = area.radii
                if area.travel_mode is not None:
                    area_dict['travel_mode'] = area.travel_mode

            elif isinstance(area, NamedArea): # named area
                area_dict = area.__studyarea__

            elif isinstance(area, dict):  # pass through - user knows what they're sending
                pass
            elif isinstance(area, list): # list of named areas, (union)
                first_area = area[0]
                ids = []
                if isinstance(first_area, NamedArea):
                    for namedarea in area:
                        a = namedarea.__studyarea__
                        if a['layer'] != first_area['layer'] or a['sourceCountry'] != first_area['sourceCountry']:
                            raise ValueError('All NamedAreas in the list must have the same source country and level')
                        ids.append(a['ids'])
                    area_dict = {"sourceCountry": first_area['sourceCountry'], "layer": first_area['layer'], "ids":[ids.join(",")]}
                else:
                    raise ValueError('Lists members must be NamedArea instances')
            else:
                raise ValueError("Don't know how to handle study areas of type " + str(type(area)))

            if comparison_levels is not None:
                # add "comparisonLevels":[{"layer": "Admin2"}, {"layer": "Admin3"}]}]
                layers = []
                for level in comparison_levels:
                    layers.append({'layer': level})

                area_dict['comparisonLevels'] = layers

            areas.append(area_dict)

    # chunking if len > 100
    if isinstance(areas, (SpatialDataFrame, pd.DataFrame, list)) and len(areas) > 100:
        parts = []
        for chunk in _chunks(l=areas, n=100):
            parts.append(ge.enrich(study_areas=chunk.copy(),
                                   data_collections=data_collections,
                                   analysis_variables=analysis_variables,
                                   add_derivative_variables=add_derivative_variables,
                                   #options=options,
                                   #use_data=use_data,
                                   intersecting_geographies=intersecting_geographies,
                                   return_geometry=return_geometry,
                                   #in_sr=in_sr,
                                   out_sr=env.out_spatial_reference,
                                   #suppress_nulls=suppress_nulls,
                                   #for_storage=for_storage,
                                   as_featureset=False))
            del chunk

        df = pd.concat(parts)
        df.reset_index(inplace=True, drop=True)
        return df
    # no chunking, len < 100, or FeatureSet
    return ge.enrich(study_areas=areas,
                      data_collections=data_collections,
                     analysis_variables=analysis_variables,
                     add_derivative_variables=add_derivative_variables,
                     #options=options,
                     #use_data=use_data,
                     intersecting_geographies=intersecting_geographies,
                     return_geometry=return_geometry,
                     #in_sr=in_sr,
                     out_sr=env.out_spatial_reference,
                     #suppress_nulls=suppress_nulls,
                     #for_storage=for_storage,
                     #as_featureset=as_featureset
                     )
#----------------------------------------------------------------------
def _find_report(country, gis=None):
    """
    Returns a list of reports by a country code

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    country                optional string. lets the user supply and optional name of a country
                           in order to get information about the data collections in that given
                           country. This should be a two country code name.
                           Example: United States as US
    ------------------     --------------------------------------------------------------------
    gis                    Optional GIS.  If None, the GIS object will be used from the
                           arcgis.env.active_gis.  This GIS object must be authenticated and
                           have the ability to consume credits
    ==================     ====================================================================

    :returns: Panda's DataFrame
    """
    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)
    return ge.find_report(country=country)
#----------------------------------------------------------------------
# def get_variables(country,
#                   dataset=None,
#                   text=None,
#                   gis=None):
#     """
#     The GeoEnrichment get_variables method allows you to search the data
#     collections for variables that contain specific keywords.
#
#     ======================     ====================================================================
#     **Argument**               **Description**
#     ----------------------     --------------------------------------------------------------------
#     country                    Optional string. Specifies the source country for the search. Use
#                                this parameter to limit the search and query of standard geographic
#                                features to one country. This parameter supports both the
#                                two-digit and three-digit country codes illustrated in the
#                                coverage table.
#
#                                Example 1 - Set source country to the United States:
#                                country=US
#
#                                Example 2 - Set source country to the Canada:
#                                country=CA
#
#                                Additional notes
#                                Currently, the service is available for Canada, the United States
#                                and a number of European countries. Other countries will be added
#                                in the near future.
#     ----------------------     --------------------------------------------------------------------
#     dataset                    optional string/list. Optional parameter to specify a specific
#                                dataset within a defined country. This parameter will not be used
#                                in the Beta release. In the future, some countries may have two or
#                                more datasets that may have different vintages and standard
#                                geography areas. For example, in the United States, there may be
#                                an optional dataset with historic census data from previous years.
#                                Examples
#                                dataset=USA_ESRI_2013
#     ----------------------     --------------------------------------------------------------------
#     text                       Optional string. Use this parameter to specify the text to query and
#                                search the data collections for the country and datasets specified.
#                                You can use this parameter to query and find specific keywords that
#                                are contained in a data collection.
#     ------------------         --------------------------------------------------------------------
#     gis                        Optional GIS.  If None, the GIS object will be used from the
#                                arcgis.env.active_gis.  This GIS object must be authenticated and
#                                have the ability to consume credits
#     ======================     ====================================================================
#
#     returns: Pandas' DataFrame
#     """
#     if gis is None:
#         gis = env.active_gis
#     ge = _GeoEnrichment(gis=gis)
#     return ge.get_variables(country=country,
#                              dataset=dataset,
#                              text=text)
#----------------------------------------------------------------------
# def report_metadata(country, gis=None):
#     """
#     This method returns information about a given country's available reports and provides
#     detailed metadata about each report.
#
#     :Usage:
#     >>> df = arcgis.geoenrichment.report_metadata("al", gis=gis)
#     # returns basic report metadata for Albania
#
#     ==================     ====================================================================
#     **Argument**           **Description**
#     ------------------     --------------------------------------------------------------------
#     country                Required string. lets the user supply and optional name of a country
#                            in order to get information about the data collections in that given
#                            country. This can be the two letter country code or the coutries
#                            full name.
#     ------------------     --------------------------------------------------------------------
#     gis                    Optional GIS.  If None, the GIS object will be used from the
#                            arcgis.env.active_gis.  This GIS object must be authenticated and
#                            have the ability to consume credits
#     ==================     ====================================================================
#
#:return: Pandas' DataFrame """ if gis is None: gis = env.active_gis ge =\
#_GeoEnrichment(gis=gis) return ge.report_metadata(country=country)\
#----------------------------------------------------------------------
@deprecated(deprecated_in="1.4.1",
            removed_in="1.5.0",
            current_version=__version__,
            details="Method will be removed due to changes in the GeoEnrichment API")
def find_businesses(type_filters=None,
                      feature_limit=1000,
                      feature_offset=0,
                      exact_match=False,
                      search_string=None,
                      spatial_filter=None,
                      simple_search=False,
                      dataset_id=None,
                      full_error_message=False,
                      out_sr=4326,
                      return_geometry=False,
                      as_featureset=False,
                      gis=None):
    """
    The find_businesses method returns business points matching a given search criteria.
    Business points can be selected using any combination of three search criteria: search
    string, spatial filter and business type. A business point will be selected if it matches
    all search criteria specified.

    ======================     ====================================================================
    **Argument**               **Description**
    ----------------------     --------------------------------------------------------------------
    type_filters               Optional list. List of business type filters restricting the search.
                               For USA, either the NAICS or SIC filter is useful as a business type
                               filter. If both filters are specified in the type_filters parameter
                               value, selected business points will match both of them.
    ----------------------     --------------------------------------------------------------------
    feature_limit              Optional integer. The limit of returned business points.
    ----------------------     --------------------------------------------------------------------
    feature_offset             Optional integer. Start the results on the number of the record
                               specified.
    ----------------------     --------------------------------------------------------------------
    exact_match                Optional boolean. True value of the parameter means the exact match
                               of the string to search.
    ----------------------     --------------------------------------------------------------------
    search_string              Optional string. A string of characters which is used in the search
                               query.
    ----------------------     --------------------------------------------------------------------
    spatial_filter             Optional SpatialFilter. A spatial filter restricting the search.
    ----------------------     --------------------------------------------------------------------
    simple_search              Optional boolean. A spatial filter restricting the search. True
                               value of the parameter means a simple search (e.g., in company
                               names only).
    ----------------------     --------------------------------------------------------------------
    dataset_id                 Optional string. ID of the active dataset.
    ----------------------     --------------------------------------------------------------------
    full_error_message         Optional boolean. Parameter for composing error message.
    ----------------------     --------------------------------------------------------------------
    out_sr                     Optional integer. Parameter specifying the spatial reference to
                               return the output dataframe.
    ----------------------     --------------------------------------------------------------------
    return_geometry            Optional boolean. When true, geometries are returned with the
                               response.
    ----------------------     --------------------------------------------------------------------
    as_featureset              Optional boolean.  The default is False. If True, the result will be
                               a arcgis.features.FeatureSet object instead of a SpatailDataFrame or
                               Pandas' DataFrame.
    ----------------------     --------------------------------------------------------------------
    gis                        Optional GIS.  If None, the GIS object will be used from the
                               arcgis.env.active_gis.  This GIS object must be authenticated and
                               have the ability to consume credits
    ======================     ====================================================================

    returns: DataFrame (Spatial or Pandas), FeatureSet, or dictionary on error.
    """
    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)
    return ge.select_businesses(type_filters=type_filters,
                                 feature_limit=feature_limit,
                                feature_offset=feature_offset,
                                exact_match=exact_match,
                                search_string=search_string,
                                spatial_filter=spatial_filter,
                                simple_search=simple_search,
                                dataset_id=dataset_id,
                                full_error_message=full_error_message,
                                out_sr=out_sr,
                                return_geometry=return_geometry,
                                as_featureset=as_featureset)
#----------------------------------------------------------------------
def standard_geography_query(source_country=None,
                             country_dataset=None,
                             layers=None,
                             ids=None,
                             geoquery=None,
                             return_sub_geography=False,
                             sub_geography_layer=None,
                             sub_geography_query=None,
                             out_sr=4326,
                             return_geometry=False,
                             return_centroids=False,
                             generalization_level=0,
                             use_fuzzy_search=False,
                             feature_limit=1000,
                             as_featureset=False,
                             gis=None):
    """
    This method allows you to search and query standard geography areas so that they can be used to
    obtain facts about the location using the enrich() method or create reports about.

    GeoEnrichment uses the concept of a study area to define the location of the point
    or area that you want to enrich with additional information. Locations can also be passed as
    one or many named statistical areas. This form of a study area lets you define an area by
    the ID of a standard geographic statistical feature, such as a census or postal area. For
    example, to obtain enrichment information for a U.S. state, county or ZIP Code or a Canadian
    province or postal code, the Standard Geography Query helper method allows you to search and
    query standard geography areas so that they can be used in the GeoEnrichment method to
    obtain facts about the location.
    The most common workflow for this service is to find a FIPS (standard geography ID) for a
    geographic name. For example, you can use this service to find the FIPS for the county of
    San Diego which is 06073. You can then use this FIPS ID within the GeoEnrichment class study
    area definition to get geometry and optional demographic data for the county. This study
    area definition is passed as a parameter to the GeoEnrichment class to return data defined
    in the enrichment pack and optionally return geometry for the feature.

    ======================     ====================================================================
    **Argument**               **Description**
    ----------------------     --------------------------------------------------------------------
    source_country             Optional string. to specify the source country for the search. Use
                               this parameter to limit the search and query of standard geographic
                               features to one country. This parameter supports both the two-digit
                               and three-digit country codes illustrated in the coverage table.
    ----------------------     --------------------------------------------------------------------
    country_dataset            Optional string. parameter to specify a specific dataset within a
                               defined country.
    ----------------------     --------------------------------------------------------------------
    layers                     Optional list/string. Parameter specifies which standard geography
                               layers are being queried or searched. If this parameter is not
                               provided, all layers within the defined country will be queried.
    ----------------------     --------------------------------------------------------------------
    ids                        Optional parameter to specify which IDs for the standard geography
                               layers are being queried or searched. You can use this parameter to
                               return attributes and/or geometry for standard geographic areas for
                               administrative areas where you already know the ID, for example, if
                               you know the Federal Information Processing Standard (FIPS) Codes for
                               a U.S. state or county; or, in Canada, to return the geometry and
                               attributes for a Forward Sortation Area (FSA).
                               Example:
                               Return the state of California where the layers parameter is set to
                               layers=['US.States']
                               then set ids=["06"]
    ----------------------     --------------------------------------------------------------------
    geoquery                   Optional string/list. This parameter specifies the text to query
                               and search the standard geography layers specified. You can use this
                               parameter to query and find standard geography features that meet an
                               input term, for example, for a list of all the U.S. counties that
                               contain the word "orange". The geoquery parameter can be a string
                               that contains one or more words.
    ----------------------     --------------------------------------------------------------------
    return_sub_geography       Optional boolean. Use this optional parameter to return all the
                               subgeographic areas that are within a parent geography.
                               For example, you could return all the U.S. counties for a given
                               U.S. state or you could return all the Canadian postal areas
                               (FSAs) within a Census Metropolitan Area (city).
                               When this parameter is set to true, the output features will be
                               defined in the sub_geography_layer. The output geometries will be
                               in the spatial reference system defined by out_sr.
    ----------------------     --------------------------------------------------------------------
    sub_geography_layer        Optional string/list. Use this optional parameter to return all the
                               subgeographic areas that are within a parent geography. For example,
                               you could return all the U.S. counties within a given U.S. state or
                               you could return all the Canadian postal areas (FSAs) within a
                               Census Metropolitan Areas (city).
                               When this parameter is set to true, the output features will be
                               defined in the sub_geography_layer. The output geometries will be
                               in the spatial reference system defined by out_sr.
    ----------------------     --------------------------------------------------------------------
    sub_geography_query        Optional string.User this parameter to filter the results of the
                               subgeography features that are returned by a search term.
                               You can use this parameter to query and find subgeography
                               features that meet an input term. This parameter is used to
                               filter the list of subgeography features that are within a
                               parent geography. For example, you may want a list of all the
                               ZIP Codes that are within "San Diego County" and filter the
                               results so that only ZIP Codes that start with "921" are
                               included in the output response. The subgeography query is a
                               string that contains one or more words.
    ----------------------     --------------------------------------------------------------------
    out_sr                     Optional integer Use this parameter to request the output geometries
                               in a specified spatial reference system.
    ----------------------     --------------------------------------------------------------------
    return_geometry            Optional boolean. Use this parameter to request the output
                               geometries in the response.  The return type will become a Spatial
                               DataFrame instead of a Panda's DataFrame.
    ----------------------     --------------------------------------------------------------------
    return_centroids           Optional Boolean.  Use this parameter to request the output geometry
                               to return the center point for each feature.
    ----------------------     --------------------------------------------------------------------
    generalization_level       Optional integer that specifies the level of generalization or
                               detail in the area representations of the administrative boundary or
                               standard geographic data layers.
                               Values must be whole integers from 0 through 6, where 0 is most
                               detailed and 6 is most generalized.
    ----------------------     --------------------------------------------------------------------
    use_fuzzy_search           Optional Boolean parameter to define if text provided in the
                               geoquery parameter should utilize fuzzy search logic. Fuzzy searches
                               are based on the Levenshtein Distance or Edit Distance algorithm.
    ----------------------     --------------------------------------------------------------------
    feature_limit              Optional integer value where you can limit the number of features
                               that are returned from the geoquery.
    ----------------------     --------------------------------------------------------------------
    as_featureset              Optional boolean.  The default is False. If True, the result will be
                               a arcgis.features.FeatureSet object instead of a SpatailDataFrame or
                               Pandas' DataFrame.
    ----------------------     --------------------------------------------------------------------
    gis                        Optional GIS.  If None, the GIS object will be used from the
                               arcgis.env.active_gis.  This GIS object must be authenticated and
                               have the ability to consume credits
    ======================     ====================================================================

    :returns: Spatial or Pandas Dataframe on success, FeatureSet, or dictionary on failure.

    """
    if gis is None:
        gis = env.active_gis
    ge = _GeoEnrichment(gis=gis)

    return ge.standard_geography_query(source_country=source_country,
                                       country_dataset=country_dataset,
                                       layers=layers,
                                       ids=ids,
                                       geoquery=geoquery,
                                       return_sub_geography=return_sub_geography,
                                       sub_geography_layer=sub_geography_layer,
                                       sub_geography_query=sub_geography_query,
                                       out_sr=out_sr,
                                       return_geometry=return_geometry,
                                       return_centroids=return_centroids,
                                       generalization_level=generalization_level,
                                       use_fuzzy_search=use_fuzzy_search,
                                       feature_limit=feature_limit,
                                       as_featureset=as_featureset)



"""
Types and functions for geocoding.
"""
import copy
from ..gis import _GISResource
import arcgis.env
import logging
from ..features import FeatureSet
from ..geometry import Geometry
_LOGGER = logging.getLogger(__name__)


class Geocoder(_GISResource):
    """
    Geocoders can find point locations of addresses, business names, and so on.
    The output points can be visualized on a map, inserted as stops for a route,
    or loaded as input for spatial analysis. It is also used to generate
    batch results for a set of addresses, as well as for reverse geocoding,
    i.e. determining the address at a particular x/y location.

    A GIS includes one or more geocoders, that can be queried using `get_geocoders(gis)`.

    Geocoders shared as Items in the GIS can be obtained using `Geocoder.fromitem(item)`.

    Geocoders may also be created using the constructor by passing in their location, such as
    a url to a Geocoding Service.
    """

    def __init__(self, location, gis=None):
        """
        Creates a Geocoder from a location, such as a url to a Geocoding Service.
        :param location: geocoder location, such as a url to a Geocoding Service.
        :param gis: the gis to which the geocoder belongs
        """
        super(Geocoder, self).__init__(location, gis)
        try:
            from arcgis.gis.server._service._adminfactory import AdminServiceGen
            self.service = AdminServiceGen(service=self, gis=gis)
        except: pass
        try:
            self._address_field = self.properties.singleLineAddressField.name
        except:
            pass #print("Geocoder does not support single line address input")

    @classmethod
    def fromitem(cls, item):
        """
        Creates a Geocoder from an Item in the GIS
        :param item: an Item of type 'Geocoding Service'
        :return: Geocoder
        """
        if not item.type == 'Geocoding Service':
            raise TypeError("item must be a type of Geocoding Service, not " + item.type)

        return cls(item.url, item._gis)

    def _geocode(self,
                 address,
                 search_extent=None,
                 location=None,
                 distance=None,
                 out_sr=None,
                 category=None,
                 out_fields="*",
                 max_locations=20,
                 magic_key=None,
                 for_storage=False,
                 as_featureset=False):
        """
        The geocode method geocodes one location per request.

        ===============     ====================================================
        **Argument**        **Description**
        ---------------     ----------------------------------------------------
        address             required list of strings or dictionaries.
                            Specifies the location to be geocoded. This can be
                            a string containing the street address, place name,
                            postal code, or POI.

                            Alternatively, this can be a dictionary containing
                            the various address fields accepted by the
                            corresponding geocoder. These fields are listed in
                            the addressFields property of the associated
                            geocoder. For example, if the address_fields of a
                            geocoder includes fields with the following names:
                            Street, City, State and Zone, then the address
                            argument is of the form:
                            {
                              Street: "1234 W Main St",
                              City: "Small Town",
                              State: "WA",
                              Zone: "99027"
                            }
        ---------------     ----------------------------------------------------
        search_extent       optional string, A set of bounding box coordinates
                            that limit the search area to a specific region.
                            This is especially useful for applications in which
                            a user will search for places and addresses only
                            within the current map extent.
        ---------------     ----------------------------------------------------
        location            optionl [x,y], Defines an origin point location that
                            is used with the distance parameter to sort
                            geocoding candidates based upon their proximity to
                            the location.
        ---------------     ----------------------------------------------------
        distance            optional float, Specifies the radius of an area
                            around a point location which is used to boost the
                            rank of geocoding candidates so that candidates
                            closest to the location are returned first. The
                            distance value is in meters.
        ---------------     ----------------------------------------------------
        out_sr              optional dictionary, The spatial reference of the
                            x/y coordinates returned by a geocode request. This
                            is useful for applications using a map with a spatial
                            reference different than that of the geocode service.
        ---------------     ----------------------------------------------------
        category            optional string, A place or address type which can
                            be used to filter find results. The parameter
                            supports input of single category values or multiple
                            comma-separated values. The category parameter can be
                            passed in a request with or without the text
                            parameter.
        ---------------     ----------------------------------------------------
        out_fields          optional string, name of all the fields to inlcude.
                            The default is "*" which means all fields.
        ---------------     ----------------------------------------------------
        max_location        optional integer, The number of locations to be
                            returned from the service. The default is 20.
        ---------------     ----------------------------------------------------
        magic_key           The find operation retrieves results quicker when
                            you pass a valid text and magickey value.
        ---------------     ----------------------------------------------------
        for_storage         Specifies whether the results of the operation will
                            be persisted. The default value is false, which
                            indicates the results of the operation can't be
                            stored, but they can be temporarily displayed on a
                            map for instance. If you store the results, in a
                            database for example, you need to set this parameter
                            to true.
        ---------------     ----------------------------------------------------
        geocoder            Optional, the geocoder to be used. If not specified,
                            the active GIS's first geocoder is used.
        ---------------     ----------------------------------------------------
        as_featureset       optional boolean, if True, the result set will be a
                            FeatureSet instead of a dictionary. False is default
        ===============     ====================================================

        :returns:
           dictionary or FeatureSet

    """
        url = self.url + "/findAddressCandidates"

        params = {
            "f": "json",
        }

        if address is not None:
            if isinstance(address, str):
                params[self._address_field] = address
            elif isinstance(address, dict):
                params.update(address)
            else:
                print("address should be a string (single line address) or dictionary "
                      "(with address fields as keys)")

        if not magic_key is None:
            params['magicKey'] = magic_key
        if not search_extent is None:
            params['searchExtent'] = search_extent
        if not location is None and \
                isinstance(location, list):
            params['location'] = "%s,%s" % (location[0], location[1])
        elif location is not None:
            params['location'] = location
        if not distance is None:
            params['distance'] = distance
        if not out_sr is None:
            params['outSr'] = out_sr
        if not category is None:
            params['category'] = category
        if out_fields is None:
            params['outFields'] = "*"
        else:
            params['outFields'] = out_fields
        if not max_locations is None:
            params['maxLocations'] = max_locations
        if not for_storage is None:
            params['forStorage'] = for_storage

        resp = self._con.post(url, params, token=self._token)

        if resp is not None and as_featureset:
            features = []
            sr = resp['spatialReference']
            for c in resp['candidates']:
                geom = c['location']
                geom['spatialReference'] = sr
                features.append({'geometry' : Geometry(geom), 'attributes' : c['attributes']})

            return FeatureSet(features=features, spatial_reference=sr)#resp['candidates']
        elif resp is not None and as_featureset == False:
            return resp['candidates']
        else:
            return []

    def _reverse_geocode(self, location, distance=None, out_sr=None, lang_code=None,
                         return_intersection=False, for_storage=False, as_featureset=False):
        """
        The reverseGeocode operation determines the address at a particular
        x/y location. You pass the coordinates of a point location to the
        geocoding service, and the service returns the address that is
        closest to the location.


        Input:
           location - a list defined as [X,Y] or a JSON Point
        """
        params = {
            "f": "json"
        }
        url = self.url + "/reverseGeocode"
        if isinstance(location, list):
            params['location'] = "%s,%s" % (location[0], location[1])
        elif isinstance(location, dict):
            params['location'] = location
        else:
            raise Exception("Invalid location")

        if distance is not None:
            params['distance'] = distance
        if out_sr is not None:
            params['outSR'] = out_sr
        if lang_code is not None:
            params['langCode'] = lang_code
        if return_intersection:
            params['returnIntersection'] = return_intersection
        if for_storage:
            params['forStorage'] = for_storage
        resp = self._con.post(url, params, token=self._token)
        if resp is not None and as_featureset:
            geom = copy.copy(resp['location'])
            del resp['location']
            fs = FeatureSet(features=[{'geometry' : Geometry(geom),
                                       "attributes" : resp['address']}])
            return fs
        return resp

    def _batch_geocode(self,
                       addresses,
                       source_country=None,
                       category=None,
                       out_sr=None,
                       as_featureset=False):
        """
        The batch_geocode() method geocodes an entire list of addresses.
        Geocoding many addresses at once is also known as bulk geocoding.


        Inputs:
           addresses - A list of addresses to be geocoded.
           For passing in the location name as a single line of text -
           single field batch geocoding - use a string.
           For passing in the location name as multiple lines of text
           multifield batch geocoding - use the address fields described
           in the Geocoder documentation.
            The maximum number of addresses that can be geocoded in a
            single request is limited to the SuggestedBatchSize property of
            the locator.
            Syntax:
             addresses = ["380 New York St, Redlands, CA",
             "1 World Way, Los Angeles, CA",
             "1200 Getty Center Drive, Los Angeles, CA",
             "5905 Wilshire Boulevard, Los Angeles, CA",
             "100 Universal City Plaza, Universal City, CA 91608",
             "4800 Oak Grove Dr, Pasadena, CA 91109"]

             OR

             addresses= [{
                "Address": "380 New York St.",
                "City": "Redlands",
                "Region": "CA",
                "Postal": "92373"
            },{
                "Address": "1 World Way",
                "City": "Los Angeles",
                "Region": "CA",
                "Postal": "90045"
            }]

           sourceCountry - The sourceCountry parameter is only supported by
            geocode services published using StreetMap Premium locators.
            Added at 10.3 and only supported by geocode services published
            with ArcGIS 10.3 for Server and later versions.
           category - The category parameter is only supported by geocode
            services published using StreetMap Premium locators.
           outSR - The well-known ID of the spatial reference, or a spatial
            reference json object for the returned addresses. For a list of
            valid WKID values, see Projected coordinate systems and
            Geographic coordinate systems.
        """
        params = {
            "f": "json"
        }
        url = self.url + "/geocodeAddresses"
        if out_sr is not None:
            params['outSR'] = out_sr
        if source_country is not None:
            params['sourceCountry'] = source_country
        if category is not None:
            params['category'] = category

        addr_recordset = []

        for index in range(len(addresses)):
            address = addresses[index]

            attributes = {"OBJECTID": index}
            if isinstance(address, str):
                attributes[self._address_field] = address
            elif isinstance(address, dict):
                attributes.update(address)
            else:
                print("Unsupported address: " + str(address))
                print("address should be a string (single line address) or dictionary "
                      "(with address fields as keys)")

            addr_rec = {"attributes": attributes}
            addr_recordset.append(addr_rec)

        params['addresses'] = {"records": addr_recordset}

        resp = self._con.post(url, params, token=self._token)
        if resp is not None and as_featureset:
            sr = resp['spatialReference']

            matches = [None] * len(addresses)
            locations = resp['locations']
            for location in locations:
                geom = copy.copy(location['location'])
                if 'spatialReference' not in geom:
                    geom['spatialReference'] = sr
                att = location['attributes']
                matches[location['attributes']['ResultID']] = {'geometry': Geometry(geom),
                                                               "attributes" : att }
            return FeatureSet(features=matches, spatial_reference=sr)
        elif resp is not None and as_featureset == False:
            matches = [None] * len(addresses)
            locations = resp['locations']
            for location in locations:
                matches[location['attributes']['ResultID']] = location
            return matches
        else:
            return []

    def _find_best_match(self,
                         address,
                         search_extent=None,
                         location=None,
                         distance=None,
                         out_sr=None,
                         category=None,
                         out_fields="*",
                         magic_key=None,
                         for_storage=False):
        """Returns the (latitude, longitude) or (y, x) coordinates of the best match for specified address"""
        candidates = self._geocode(address, search_extent, location, distance,
                                   out_sr, category, out_fields, 1, magic_key, for_storage)
        if candidates:
            location = candidates[0]['location']
            return location['y'], location['x']

    def _suggest(self,
                 text,
                 location,
                 distance=None,
                 category=None):
        """
        The suggest operation is performed on a geocoder.
        The result of this operation is a resource representing a list of
        suggested matches for the input text. This resource provides the
        matching text as well as a unique ID value, which links a
        suggestion to a specific place or address.
        A geocode service must meet the following requirements to support
        the suggest operation:
          The address locator from which the geocode service was published
          must support suggestions. Only address locators created using
          ArcGIS 10.3 for Desktop and later can support suggestions. See
          the Create Address Locator geoprocessing tool help topic for more
          information.
          The geocode service must have the Suggest capability enabled.
          Only geocode services published using ArcGIS 10.3 for Server or
          later support the Suggest capability.
        The suggest operation allows character-by-character auto-complete
        suggestions to be generated for user input in a client application.
        This capability facilitates the interactive search user experience
        by reducing the number of characters that need to be typed before
        a suggested match is obtained. A client application can provide a
        list of suggestions that is updated with each character typed by a
        user until the address they are looking for appears in the list.
        Inputs:
           text - The input text provided by a user that is used by the
            suggest operation to generate a list of possible matches. This
            is a required parameter.
           location -  Defines an origin point location that is used with
            the distance parameter to sort suggested candidates based on
            their proximity to the location. The distance parameter
            specifies the radial distance from the location in meters. The
            priority of candidates within this radius is boosted relative
            to those outside the radius.
            This is useful in mobile applications where a user wants to
            search for places in the vicinity of their current GPS
            location. It is also useful for web mapping applications where
            a user wants to find places within or near the map extent.
            The location parameter can be specified without specifying a
            distance. If distance is not specified, it defaults to 2000
            meters.
            The object can be an common.geometry.Point or X/Y list object
           distance - Specifies the radius around the point defined in the
            location parameter to create an area, which is used to boost
            the rank of suggested candidates so that candidates closest to
            the location are returned first. The distance value is in
            meters.
            If the distance parameter is specified, the location parameter
            must be specified as well.
            It is important to note that the location and distance
            parameters allow searches to extend beyond the specified search
            radius. They are not used to filter results, but rather to rank
            resulting candidates based on their distance from a location.
           category - The category parameter is only supported by geocode
            services published using StreetMap Premium locators.
        """
        params = {
            "f": "json",
            "text": text
        }
        url = self.url + "/suggest"

        if isinstance(location, list):
            params['location'] = "%s,%s" % (location[0], location[1])
        else:
            raise Exception("Invalid location, please try again")
        if not category is None:
            params['category'] = category
        if not distance is None and \
                isinstance(distance, (int, float)):
            params['distance'] = distance
        resp = self._con.post(url, params, token=self._token)
        return resp


def get_geocoders(gis):
    """
    A GIS includes one or more geocoders. The list of geocoders registered with the GIS
    can be queried using get_geocoders.

    :param gis: the GIS whose registered geocoders are to be queried


    :return: list of geocoders registered with the GIS
    """
    geocoders = []
    try:
        geocode_services = gis.properties['helperServices']['geocode']
        for geocode_service in geocode_services:
            try:
                geocoders.append(Geocoder(geocode_service['url'], gis))
            except RuntimeError as runtime_error:
                _LOGGER.warning('Unable to use Geocoder at ' + geocode_service['url'])
                _LOGGER.warning(str(runtime_error))
    except KeyError:
        pass
    return geocoders

#----------------------------------------------------------------------
def analyze_geocode_input(input_table_or_item,
                          geocode_service_url=None,
                          column_names=None,
                          input_file_parameters=None,
                          locale="en",
                          context=None,
                          gis=None):
    """
    The analyze_geocode_input function takes in a geocode input (either a table or file of
    addresses) and returns an output dictionary that includes a suggested field mapping. It supports CSV,
    XLS, or table input. The table can be from a big data file share or from a feature service. The
    task generates a suggested field mapping based on the input fields and the geocoding service
    candidate fields and returns it in a geocode_parameters dictionary. This geocode_parameters
    dictionary output is the an input to the Batch Geocode tool. The output geocode_parameters
    dictionary also includes field info (name, length, and type) as well as additional information
    that helps the geocode tool parse the input file or table.

    =====================     ================================================================
    **Argument**              **Description**
    ---------------------     ----------------------------------------------------------------
    input_table_or_item       required Item, string or dictionary.  The input to analyze for
                              geocoding.

                              For tables:

                               The input table specification must include the following:

                               - A URL to an input table
                               - A service token to access the table
                               Note that if the table is a hosted table on the same portal,
                               serviceToken is not required.

                               Example: {"url":"<table url>","serviceToken":"<token>"}

                              For File Items:

                              The input file should be a portal item. Input the itemid of
                              the item in the portal. The format of the item in the portal
                              can be in one of the following formats:

                              - CSV
                              - Microsoft Excel spreadsheet (XLSX)
                              Example: {"itemid": "<itemid of file>" }
    ---------------------     ----------------------------------------------------------------
    geocode_service_url       Optional string or Geocode class.  The geocode service that you
                              want to geocode your addresses against.
    ---------------------     ----------------------------------------------------------------
    column_names              Optional string.  Only used when input table or Item has no
                              header row.
                              Example: address,city,state,zip
    ---------------------     ----------------------------------------------------------------
    input_file_parameters     Optional dictionary. Enter information about how to parse the
                              file. If you are using an input table instead of an Item as
                              input, this parameter can be left blank.
                              Any of the key values in the dictionary can be left blank using
                              the "".

                              Values:

                              fileType - Enter CSV or XLS for the file format of file Item.
                              headerRowExists - Enter true if your file has a header row,
                                                false if it does not.
                              columnDelimiter - Enter SPACE, TAB, COMMA, PIPE, or SEMICOLON.
                              textQualifier - Enter either SINGLE_QUOTE or DOUBLE_QUOTE.

                              Example: {"fileType":"xlsx","headerRowExists":"true",
                                        "columnDelimiter":"","textQualifier":""}
    ---------------------     ----------------------------------------------------------------
    locale                    Optional string. Enter the 2-letter ("en") or 4-letter ("ar-il")
                              specific locale if geocodeInput is in a language other than
                              English.
    ---------------------     ----------------------------------------------------------------
    context                   Optional dictionary.
                              Context contains additional settings that affect task execution.
                              Analyze_geocode_input() has the following two settings:
                              1. Extent (extent) - A bounding box that defines the analysis
                                 area. Only those points in inputLayer that intersect the
                                 bounding box are analyzed.
                              2. Output Spatial Reference (outSR) - The output features are
                                 projected into the output spatial reference.
    ---------------------     ----------------------------------------------------------------
    gis                       Optional GIS. Connection to the site. If None is given, the
                              active GIS is used.
    =====================     ================================================================

    :returns: dictionary

    :Usage Example:

    >>>res = analyze_geocode_input(geocode_service_url=Geocoder,
                                   input_table_or_item={"itemid" : "abc123545asv"},
                                   input_file_parameters={"fileType":"csv","headerRowExists":"true",
                                                         "columnDelimiter":"","textQualifier":""})
    >>> print(res)
    {'header_row_exists': True, 'field_info': '[["Address", "TEXT", 255], ["City", "TEXT", 255],
    ["State", "TEXT", 255], ["ZipCode", "TEXT", 255]]', 'file_type': 'csv', 'field_mapping': '[["Address", ""],
    ["City", "City"], ["State", "State"], ["ZipCode", ""]]', 'column_names': '',
    'column_delimiter': '', 'text_qualifier': '', 'singleline_field': 'Single Line Input'}

    :Usage Example 2:
    >>> table_lyr = Table(url="http://testsite.com/server/rest/services/Hosted/addresses/FeatureServer/0", gis=gis)
    >>> res = analyze_geocode_input()
    """
    import json
    from arcgis.gis import Item
    from arcgis.features.layer import Layer
    from arcgis.geoprocessing._tool import Toolbox

    gis = arcgis.env.active_gis if gis is None else gis
    analyze_geocode_url = gis.properties.helperServices.asyncGeocode.url
    tbx = Toolbox(url=analyze_geocode_url, gis=gis)

    if geocode_service_url is None:
        gcs = gis.properties.helperServices.geocode
        for gc in gcs:
            if 'batch' in gc and gc['batch']:
                geocode_service_url = gc['url']
                break
            del gc
        del gcs
    elif isinstance(geocode_service_url, Geocoder):
        geocode_service_url = geocode_service_url.url
    elif isinstance(geocode_service_url, str) == False:
        raise ValueError("Invalid geocoder service given.")
    if geocode_service_url is None:
        raise ValueError("The registered geocoders are not valid to use with this tool.")

    kwargs = {"geocode_service_url" : geocode_service_url,
              "input_table" : "",
              "input_file_item" : None,
              "column_names" : column_names,
              "input_file_parameters" : input_file_parameters,
              "locale" : locale,
              "context" : context}


    if isinstance(input_table_or_item, Item):
        kwargs['input_file_item'] = {'itemid' : input_table_or_item.itemid}
        if input_file_parameters is None:
            kwargs['input_file_parameters'] = json.dumps({"fileType": input_table_or_item.type.lower(),
                                               "headerRowExists":"true",
                                               "columnDelimiter":"",
                                               "textQualifier":""})
    elif isinstance(input_table_or_item, str):
        kwargs['input_file_item'] = {'itemid' : input_table_or_item}
        if input_file_parameters is None:
            item = gis.content.search('id: %s' % input_table_or_item)[0]
            kwargs['input_file_parameters'] = json.dumps({"fileType": item.type.lower(),
                                                          "headerRowExists":"true",
                                                          "columnDelimiter":"",
                                                          "textQualifier":""})
    elif isinstance(input_table_or_item, dict):
        if "url" in input_table_or_item:
            kwargs['input_table'] = input_table_or_item
        elif "itemid" in input_table_or_item:
            kwargs['input_file_item'] = input_table_or_item
            if input_file_parameters is None:
                item = gis.content.search('id: %s' % input_table_or_item['itemid'])[0]
                kwargs['input_file_parameters'] = json.dumps({"fileType": item.type.lower(),
                                                              "headerRowExists":"true",
                                                              "columnDelimiter":"",
                                                              "textQualifier":""})
    elif isinstance(input_table_or_item, Layer):
        lyr_dict = input_table_or_item._lyr_dict
        if 'type' in lyr_dict:
            lyr_dict.pop("type")
        kwargs['input_table'] = lyr_dict
    for k,v in list(kwargs.items()):
        if v is None:
            kwargs.pop(k)
    return tbx.analyze_geocode_input(**kwargs)

#----------------------------------------------------------------------
def geocode_from_items(input_data,
                       output_type='Feature Layer',
                       geocode_service_url=None,
                       geocode_parameters=None,
                       country=None,
                       output_fields=None,
                       header_rows_to_skip=1,
                       output_name=None,
                       category=None,
                       context=None,
                       gis=None
                       ):
    """
    The Batch Geocode geocodes a table or file of addresses and returns the geocoded results. It
    supports CSV, XLS or table input. The task geocodes the entire file regardless of size.

    =====================     ================================================================
    **Argument**              **Description**
    ---------------------     ----------------------------------------------------------------
    input_data                required Item, string, Layer. Data to geocode.
    ---------------------     ----------------------------------------------------------------
    output_type               optional string.  Export item types.  Allowed values are CSV,
                              XLS, or Feature Layer (default)
    ---------------------     ----------------------------------------------------------------
    geocode_service_url       optional string of Geocoder. Optional geocoder to use to
                              spatially enable the dataset.
    ---------------------     ----------------------------------------------------------------
    geocode_parameters        optional dictionary.  This includes parameters that help parse
                              the input data, as well the field lengths and a field mapping.
                              This value is the output from the analyze_geocode_input()
                              available on your server designated to geocode. It is important
                              to inspect the field mapping closely and adjust them accordingly
                              before submitting your job, otherwise your geocoding results may
                              not be accurate. It is recommended to use the output from
                              analyze_geocode_input() and modify the field mapping instead of
                              constructing this dictionary by hand.

                              **Values**

                              **field_info** - A list of triples with the field names of your input
                              data, the field type (usually TEXT), and the allowed length
                              (usually 255).
                              Example: [['ObjectID', 'TEXT', 255], ['Address', 'TEXT', 255],
                                       ['Region', 'TEXT', 255], ['Postal', 'TEXT', 255]]

                              **header_row_exists** - Enter true or false.

                              **column_names** - Submit the column names of your data if your data
                              does not have a header row.

                              **field_mapping** - Field mapping between each input field and
                              candidate fields on the geocoding service.
                              Example: [['ObjectID', 'OBJECTID'], ['Address', 'Address'],
                                          ['Region', 'Region'], ['Postal', 'Postal']]
    ---------------------     ----------------------------------------------------------------
    country                   optional string.  If all your data is in one country, this helps
                              improve performance for locators that accept that variable.
    ---------------------     ----------------------------------------------------------------
    output_fields             optional string. Enter the output fields from the geocoding
                              service that you want returned in the results, separated by
                              commas. To output all available outputFields, leave this
                              parameter blank.
                              Example: score,match_addr,x,y
    ---------------------     ----------------------------------------------------------------
    header_rows_to_skip       optional integer. Describes on which row your data begins in
                              your file or table. The default is 1 (since the first row
                              contains the headers). The default is 1.
    ---------------------     ----------------------------------------------------------------
    output_name               optional string, The task will create a feature service of the
                              results. You define the name of the service.
    ---------------------     ----------------------------------------------------------------
    category                  optional string. Enter a category for more precise geocoding
                              results, if applicable. Some geocoding services do not support
                              category, and the available options depend on your geocode service.
    ---------------------     ----------------------------------------------------------------
    context                   optional dictionary. Context contains additional settings that
                              affect task execution. Batch Geocode has the following two
                              settings:

                              1. Extent (extent) - A bounding box that defines the analysis
                                 area. Only those points in inputLayer that intersect the
                                 bounding box are analyzed.
                              2. Output Spatial Reference (outSR) - The output features are
                                 projected into the output spatial reference.
                              Syntax:
                              {
                              "extent" : {extent}
                              "outSR" : {spatial reference}
                              }
    ---------------------     ----------------------------------------------------------------
    gis                       optional GIS, the GIS on which this tool runs. If not specified,
                              the active GIS is used.
    =====================     ================================================================


    :returns: arcgis.gis.Item
    """

    import json
    import uuid
    from arcgis.gis import Item
    from arcgis.geocoding import Geocoder
    from arcgis.features.layer import Layer
    from arcgis.geoprocessing._tool import Toolbox
    from arcgis.geoanalytics._util import _create_output_service

    uid = uuid.uuid4().hex[:5]
    locator_parameters = None

    kwargs = {
        "geocode_parameters" : geocode_parameters,
        "geocode_service_url" : geocode_service_url,
        "output_type" : output_type,
        "input_table" : "",
        "input_file_item" : None,
        "source_country" : country,
        "category" : category,
        "output_fields" : output_fields,
        "header_rows_to_skip" : header_rows_to_skip,
        "output_name" : output_name,
        "context" : context,
        "locator_parameters" : locator_parameters
    }

    _item_type = "SERVICE"
    item = None
    url = gis.properties.helperServices.asyncGeocode.url
    tbx = Toolbox(url=url, gis=gis)

    if output_type is None:
        output_type = "CSV"
    if output_type.lower() == 'xlsx':
        output_type = 'XLS'
    if output_type.lower() == "feature layer":
        output_type = "Feature Service"
    if output_type not in ['CSV', 'XLS', "Feature Service"]:
        raise ValueError("Invalid output_type: %s" % output_type)

    if geocode_service_url is None:
        gcs = gis.properties.helperServices.geocode
        for gc in gcs:
            if 'batch' in gc and gc['batch']:
                geocode_service_url = gc['url']
                kwargs['geocode_service_url'] = gc['url']
                break
            del gc
        del gcs
    elif isinstance(geocode_service_url, Geocoder):
        geocode_service_url = geocode_service_url.url
        kwargs['geocode_service_url'] = geocode_service_url.url
    elif isinstance(geocode_service_url, str) == False:
        raise ValueError("Invalid geocoder service given.")
    if geocode_service_url is None:
        raise ValueError("The registered geocoders are not valid to use with this tool.")

    if isinstance(input_data, Item):
        _item_type = input_data.type
        kwargs['input_file_item'] = {'itemid' : input_data.itemid}
    elif isinstance(input_data, dict):
        if 'url' in input_data:
            _item_type = "SERVICE"
            kwargs['input_table'] = input_data
        elif 'itemid' in input_data:
            item = gis.content.search("id: %s" % input_data['itemid'])[0]
            _item_type = item.type
            kwargs['input_file_item'] = input_data
        pass
    elif isinstance(input_data, str):
        item = gis.content.search("id: %s" % input_data)[0]
        _item_type = item.type
        kwargs['input_file_item'] = {'itemid' : input_data}
    elif isinstance(input_data, Layer):
        lyr = input_data._lyr_dict
        if 'type' in lyr:
            del lyr['type']
        kwargs['input_table'] = lyr
    else:
        raise ValueError("Invalid input_data")

    # Figure out input_data
    if geocode_parameters is None and \
       _item_type in ['CSV', 'csv', 'XLS',
                      'xls', 'XLSX', 'xlsx']:
        import json
        if header_rows_to_skip is None:
            hre = "false"
        else:
            hre = "true"
        kwargs['geocode_parameters'] = analyze_geocode_input(
            input_table_or_item=kwargs['input_file_item'],
            geocode_service_url=geocode_service_url,
            gis=gis)

    if output_type == "Feature Layer" and \
       output_name is None:
        if output_name is None:
            output_service = _create_output_service(gis=gis,
                               output_name='Geocoded_Feature_Service_ %' % uid,
                              output_service_name='Geocoded_Feature_Service_ %' % uid,
                              task='Geocoding')
        else:#output_type == 'Feature Layer':
            output_service = _create_output_service(gis=gis,
                                     output_name='Geocoded_Feature_Service_%s' % uid,
                                     output_service_name='Geocoded_Feature_Service_%s' % uid,
                                     task='Geocoding')
        kwargs['output_name'] = json.dumps({
            "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
            "itemProperties": {"itemId" : output_service.itemid}})
    elif output_type in ['XLS', 'xls', 'XLSX', 'xlsx']:
        if output_name:
            kwargs['output_name'] = {"itemProperties":{"title":"Geocoded Results %s" % output_name,
                                                   "description":"Geocoded results for %s generated from running the Geocode Locations from Table solution." % output_name,
                                                   "tags":"Analysis Result, Geocode Locations From Table",
                                                   "snippet":"Excel File generated from Geocode Locations From Table","folderId":""}}
        else:
            output_name = 'Geocoded_Result_ %' % uid
            kwargs['output_name'] = {"itemProperties":{"title":"Geocoded Results %s" % output_name,
                                                       "description":"Geocoded results for %s generated from running the Geocode Locations from Table solution." % output_name,
                                                       "tags":"Analysis Result, Geocode Locations From Table",
                                                       "snippet":"Excel File generated from Geocode Locations From Table","folderId":""}}
    elif output_type in ['CSV', 'csv']:
        if output_name:
            kwargs['output_name'] = {"itemProperties":{"title":"Geocoded Results %s" % output_name,
                                                   "description":"Geocoded results for %s generated from running the Geocode Locations from Table solution." % output_name,
                                                   "tags":"Analysis Result, Geocode Locations From Table",
                                                   "snippet":"CSV File generated from Geocode Locations From Table","folderId":""}}
        else:
            output_name = 'Geocoded_Result_%s' % uid
            kwargs['output_name'] = {"itemProperties":{"title":"Geocoded Results %s" % output_name,
                                                       "description":"Geocoded results for %s generated from running the Geocode Locations from Table solution." % output_name,
                                                       "tags":"Analysis Result, Geocode Locations From Table",
                                                       "snippet":"CSV File generated from Geocode Locations From Table","folderId":""}}

    for k,v in list(kwargs.items()):
        if v is None:
            kwargs.pop(k)
    res = tbx.batch_geocode(**kwargs)
    if 'itemId' in res:
        return gis.content.get(res['itemId'])
    else:
        return res

def geocode(address,
            search_extent=None,
            location=None,
            distance=None,
            out_sr=None,
            category=None,
            out_fields="*",
            max_locations=20,
            magic_key=None,
            for_storage=False,
            geocoder=None,
            as_featureset=False):
    """
    The geocode function geocodes one location per request.

    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    address             required list of strings or dictionaries.
                        Specifies the location to be geocoded. This can be
                        a string containing the street address, place name,
                        postal code, or POI.

                        Alternatively, this can be a dictionary containing
                        the various address fields accepted by the
                        corresponding geocoder. These fields are listed in
                        the addressFields property of the associated
                        geocoder. For example, if the address_fields of a
                        geocoder includes fields with the following names:
                        Street, City, State and Zone, then the address
                        argument is of the form:
                        {
                          Street: "1234 W Main St",
                          City: "Small Town",
                          State: "WA",
                          Zone: "99027"
                        }
    ---------------     ----------------------------------------------------
    search_extent       optional string, A set of bounding box coordinates
                        that limit the search area to a specific region.
                        This is especially useful for applications in which
                        a user will search for places and addresses only
                        within the current map extent.
    ---------------     ----------------------------------------------------
    location            optionl [x,y], Defines an origin point location that
                        is used with the distance parameter to sort
                        geocoding candidates based upon their proximity to
                        the location.
    ---------------     ----------------------------------------------------
    distance            optional float, Specifies the radius of an area
                        around a point location which is used to boost the
                        rank of geocoding candidates so that candidates
                        closest to the location are returned first. The
                        distance value is in meters.
    ---------------     ----------------------------------------------------
    out_sr              optional dictionary, The spatial reference of the
                        x/y coordinates returned by a geocode request. This
                        is useful for applications using a map with a spatial
                        reference different than that of the geocode service.
    ---------------     ----------------------------------------------------
    category            optional string, A place or address type which can
                        be used to filter find results. The parameter
                        supports input of single category values or multiple
                        comma-separated values. The category parameter can be
                        passed in a request with or without the text
                        parameter.
    ---------------     ----------------------------------------------------
    out_fields          optional string, name of all the fields to inlcude.
                        The default is "*" which means all fields.
    ---------------     ----------------------------------------------------
    max_location        optional integer, The number of locations to be
                        returned from the service. The default is 20.
    ---------------     ----------------------------------------------------
    magic_key           The find operation retrieves results quicker when
                        you pass a valid text and magickey value.
    ---------------     ----------------------------------------------------
    for_storage         Specifies whether the results of the operation will
                        be persisted. The default value is false, which
                        indicates the results of the operation can't be
                        stored, but they can be temporarily displayed on a
                        map for instance. If you store the results, in a
                        database for example, you need to set this parameter
                        to true.
    ---------------     ----------------------------------------------------
    geocoder            Optional, the geocoder to be used. If not specified,
                        the active GIS's first geocoder is used.
    ---------------     ----------------------------------------------------
    as_featureset       optional boolean, if True, the result set is
                        returned as a FeatureSet object, else it is a
                        dictionary.
    ===============     ====================================================

    :returns:
       dictionary

    """
    # as_featureset = False
    if geocoder is None:
        geocoder = arcgis.env.active_gis._tools.geocoders[0]
    return geocoder._geocode(
        address,
        search_extent,
        location,
        distance,
        out_sr,
        category,
        out_fields,
        max_locations,
        magic_key,
        for_storage,
        as_featureset)


def reverse_geocode(location, distance=None, out_sr=None, lang_code=None,
                    return_intersection=False,
                    for_storage=False, geocoder=None):
    """
    The reverse_geocode operation determines the address at a particular
    x/y location. You pass the coordinates of a point location to the
    geocoding service, and the service returns the address that is
    closest to the location.

    =================== ====================================================
    **Argument**        **Description**
    ------------------- ----------------------------------------------------
    location            required list/Point Geometry
    ------------------- ----------------------------------------------------
    distance            optional float, radial distance in meteres to
                        search for an address.  The default is 100 meters.
    ------------------- ----------------------------------------------------
    out_sr              optional integer, spatial reference of the x/y
                        coordinate returned.
    ------------------- ----------------------------------------------------
    return_intersection optional Boolean, which specifies whether the
                        service should return the nearest street
                        intersection or the nearest address to the input
                        location
    ------------------- ----------------------------------------------------
    for_storage         optional boolean, specifies whether the results of
                        the operation will be persisted
    ------------------- ----------------------------------------------------
    geocoder            optional geocoder, the geocoder to be used. If not
                        specified, the active GIS's first geocoder is used.
    =================== ====================================================

    :returns:
       dictionary
    """

    if geocoder is None:
        geocoder = arcgis.env.active_gis._tools.geocoders[0]
    return geocoder._reverse_geocode(location, distance, out_sr, lang_code,
                                     return_intersection,
                                     for_storage)


def batch_geocode(addresses,
                  source_country=None,
                  category=None,
                  out_sr=None,
                  geocoder=None,
                  as_featureset=False):
    """
    The batch_geocode() function geocodes an entire list of addresses.
    Geocoding many addresses at once is also known as bulk geocoding.

    ===============     ================================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------------------
    addresses           required list of strings or dictionaries.
                        A list of addresses to be geocoded.
                        For passing in the location name as a single line of text -
                        single field batch geocoding - use a string.
                        For passing in the location name as multiple lines of text
                        multifield batch geocoding - use the address fields described
                        in the Geocoder documentation.
                         The maximum number of addresses that can be geocoded in a
                         single request is limited to the SuggestedBatchSize property of
                         the locator.
                         Syntax:
                          addresses = ["380 New York St, Redlands, CA",
                          "1 World Way, Los Angeles, CA",
                          "1200 Getty Center Drive, Los Angeles, CA",
                          "5905 Wilshire Boulevard, Los Angeles, CA",
                          "100 Universal City Plaza, Universal City, CA 91608",
                          "4800 Oak Grove Dr, Pasadena, CA 91109"]

                          OR

                          addresses= [{
                             "Address": "380 New York St.",
                             "City": "Redlands",
                             "Region": "CA",
                             "Postal": "92373"
                         },{
                             "Address": "1 World Way",
                             "City": "Los Angeles",
                             "Region": "CA",
                             "Postal": "90045"
                         }]
    ---------------     ----------------------------------------------------------------
    source_country      optional string, The source_country parameter is
                        only supported by geocoders published using StreetMap
                        Premium locators.
                        Added at 10.3 and only supported by geocoders published
                        with ArcGIS 10.3 for Server and later versions.
    ---------------     ----------------------------------------------------------------
    category            The category parameter is only supported by geocode
                        services published using StreetMap Premium locators.
    ---------------     ----------------------------------------------------------------
    out_sr              optional dictionary, The spatial reference of the
                        x/y coordinates returned by a geocode request. This
                        is useful for applications using a map with a spatial
                        reference different than that of the geocode service.
    ---------------     ----------------------------------------------------------------
    as_featureset       optional boolean, if True, the result set is
                        returned as a FeatureSet object, else it is a
                        dictionary.
    ---------------     ----------------------------------------------------------------
    geocoder            Optional, the geocoder to be used. If not specified,
                        the active GIS's first geocoder is used.
    ===============     ================================================================

    :returns:
       dictionary or FeatureSet
    """
    if geocoder is None:
        geocoder = arcgis.env.active_gis._tools.geocoders[0]
    return geocoder._batch_geocode(
        addresses,
        source_country,
        category,
        out_sr,
        as_featureset)


def suggest(text,
            location,
            distance=None,
            category=None,
            geocoder=None):
    """
    The result of this operation is a resource representing a list of
    suggested matches for the input text. This resource provides the
    matching text as well as a unique ID value, which links a
    suggestion to a specific place or address.
    A geocoder must meet the following requirements to support
    the suggest operation:
      The address locator from which the geocoder was published
      must support suggestions. Only address locators created using
      ArcGIS 10.3 for Desktop and later can support suggestions. See
      the Create Address Locator geoprocessing tool help topic for more
      information.
      The geocoder must have the Suggest capability enabled.
      Only geocoders published using ArcGIS 10.3 for Server or
      later support the Suggest capability.
    The suggest operation allows character-by-character auto-complete
    suggestions to be generated for user input in a client application.
    This capability facilitates the interactive search user experience
    by reducing the number of characters that need to be typed before
    a suggested match is obtained. A client application can provide a
    list of suggestions that is updated with each character typed by a
    user until the address they are looking for appears in the list.

    ===============     =================================================================
    **Argument**        **Description**
    ---------------     -----------------------------------------------------------------
    text                The input text provided by a user that is used by the
                        suggest operation to generate a list of possible
                        matches. This is a required parameter.
    ---------------     -----------------------------------------------------------------
    location            Defines an origin point location that is used with
                        the distance parameter to sort suggested candidates
                        based on their proximity to the location. The
                        distance parameter specifies the radial distance from
                        the location in meters. The priority of candidates
                        within this radius is boosted relative to those
                        outside the radius.
                        This is useful in mobile applications where a user
                        wants to search for places in the vicinity of their
                        current GPS location. It is also useful for web
                        mapping applications where a user wants to find
                        places within or near the map extent.
                        The location parameter can be specified without
                        specifying a distance. If distance is not specified,
                        it defaults to 2000 meters.
    ---------------     -----------------------------------------------------------------
    distance            Specifies the radius around the point defined in the
                        location parameter to create an area, which is used to boost
                        the rank of suggested candidates so that candidates closest to
                        the location are returned first. The distance value is in
                        meters.
                        If the distance parameter is specified, the location parameter
                        must be specified as well.
                        It is important to note that the location and distance
                        parameters allow searches to extend beyond the specified search
                        radius. They are not used to filter results, but rather to rank
                        resulting candidates based on their distance from a location.
    ---------------     -----------------------------------------------------------------
    category            The category parameter is only supported by geocode
                        services published using StreetMap Premium locators.
    ---------------     -----------------------------------------------------------------
    geocoder            Optional, the geocoder to be used. If not specified,
                        the active GIS's first geocoder is used.
    ===============     =================================================================
    """
    if geocoder is None:
        geocoder = arcgis.env.active_gis._tools.geocoders[0]
    return geocoder._suggest(
        text,
        location,
        distance,
        category)



"""
This module provides built-in decorators that can be used in order to allow for unified handling of
deprecation and warning notications to the users.  The module provides a simple easy to use '@'
notation with some required parameters to given unified deprecation warnings/exceptions.

Module is a fork of project: https://github.com/briancurtin/deprecation under Apache 2.0 License
"""
import collections
import functools
import textwrap
import warnings
import re

__version__ = "2.0.0"

# This is mostly here so automodule docs are ordered more ideally.
__all__ = ["deprecated", "message_location",
           "DeprecatedWarning", "UnsupportedWarning"]

# message_location provides some simple
message_location = "top"



def _parse(version):
    def normalize(v):
        import re
        return [int(x) for x in re.sub(r'(\.0+)*$','', v).split(".")]
    return normalize(version)

class DeprecatedWarning(DeprecationWarning):
    """A warning class for deprecated methods

    This is a specialization of the built-in :class:`DeprecationWarning`,
    adding parameters that allow us to get information into the __str__
    that ends up being sent through the :mod:`warnings` system.
    The attributes aren't able to be retrieved after the warning gets
    raised and passed through the system as only the class--not the
    instance--and message are what gets preserved.

    :param function: The function being deprecated.
    :param deprecated_in: The version that ``function`` is deprecated in
    :param removed_in: The version that ``function`` gets removed in
    :param details: Optional details about the deprecation. Most often
                    this will include directions on what to use instead
                    of the now deprecated code.
    """

    def __init__(self, function, deprecated_in, removed_in, details=""):
        self.function = function
        self.deprecated_in = deprecated_in
        self.removed_in = removed_in
        self.details = details
        super(DeprecatedWarning, self).__init__()

    def __str__(self):
        # Use a defaultdict to give us the empty string
        # when a part isn't included.
        parts = collections.defaultdict(str)
        parts["function"] = self.function

        if self.deprecated_in:
            parts["deprecated"] = " as of %s" % self.deprecated_in
        if self.removed_in:
            parts["removed"] = " and will be removed in %s" % self.removed_in
        if any([self.deprecated_in, self.removed_in, self.details]):
            parts["period"] = "."
        if self.details:
            parts["details"] = " %s" % self.details

        return ("%(function)s is deprecated%(deprecated)s%(removed)s"
                "%(period)s%(details)s" % (parts))


class UnsupportedWarning(DeprecatedWarning):
    """A warning class for methods to be removed

    This is a subclass of :class:`deprecation.DeprecatedWarning` and is used
    to output a proper message about a function being unsupported.
    Additionally, the :func:`deprecation.fail_if_not_removed` decorator
    will handle this warning and cause any tests to fail if the system
    under test uses code that raises this warning.
    """

    def __str__(self):
        parts = collections.defaultdict(str)
        parts["function"] = self.function
        parts["removed"] = self.removed_in

        if self.details:
            parts["details"] = " %s" % self.details

        return ("%(function)s is unsupported as of %(removed)s."
                "%(details)s" % (parts))

def deprecated(deprecated_in=None, removed_in=None, current_version=None,
               details=""):
    """Decorate a function to signify its deprecation

    This function wraps a method that will soon be removed and does two things:
        * The docstring of the method will be modified to include a notice
          about deprecation, e.g., "Deprecated since 0.9.11. Use foo instead."
        * Raises a :class:`~deprecation.DeprecatedWarning`
          via the :mod:`warnings` module, which is a subclass of the built-in
          :class:`DeprecationWarning`. Note that built-in
          :class:`DeprecationWarning`\s are ignored by default, so for users
          to be informed of said warnings they will need to enable them--see
          the :mod:`warnings` module documentation for more details.

    :param deprecated_in: The version at which the decorated method is
                          considered deprecated. This will usually be the
                          next version to be released when the decorator is
                          added. The default is **None**, which effectively
                          means immediate deprecation. If this is not
                          specified, then the `removed_in` and
                          `current_version` arguments are ignored.
    :param removed_in: The version when the decorated method will be removed.
                       The default is **None**, specifying that the function
                       is not currently planned to be removed.
                       Note: This cannot be set to a value if
                       `deprecated_in=None`.
    :param current_version: The source of version information for the
                            currently running code. This will usually be
                            a `__version__` attribute on your library.
                            The default is `None`.
                            When `current_version=None` the automation to
                            determine if the wrapped function is actually
                            in a period of deprecation or time for removal
                            does not work, causing a
                            :class:`~deprecation.DeprecatedWarning`
                            to be raised in all cases.
    :param details: Extra details to be added to the method docstring and
                    warning. For example, the details may point users to
                    a replacement method, such as "Use the foo_bar
                    method instead". By default there are no details.
    """
    # You can't just jump to removal. It's weird, unfair, and also makes
    # building up the docstring weird.
    if deprecated_in is None and removed_in is not None:
        raise TypeError("Cannot set removed_in to a value "
                        "without also setting deprecated_in")

    # Only warn when it's appropriate. There may be cases when it makes sense
    # to add this decorator before a formal deprecation period begins.
    # In CPython, PendingDeprecatedWarning gets used in that period,
    # so perhaps mimick that at some point.
    is_deprecated = False
    is_unsupported = False

    # StrictVersion won't take a None or a "", so make whatever goes to it
    # is at least *something*.
    if current_version:
        current_version = _parse(current_version)

        if (removed_in
            and current_version >= _parse(removed_in)):
            is_unsupported = True
        elif (deprecated_in
              and current_version >= _parse(deprecated_in)):
            is_deprecated = True
    else:
        # If we can't actually calculate that we're in a period of
        # deprecation...well, they used the decorator, so it's deprecated.
        # This will cover the case of someone just using
        # @deprecated("1.0") without the other advantages.
        is_deprecated = True

    should_warn = any([is_deprecated, is_unsupported])

    def _function_wrapper(function):
        if should_warn:
            # Everything *should* have a docstring, but just in case...
            existing_docstring = function.__doc__ or ""

            # The various parts of this decorator being optional makes for
            # a number of ways the deprecation notice could go. The following
            # makes for a nicely constructed sentence with or without any
            # of the parts.
            parts = {
                "deprecated_in":
                    " %s" % deprecated_in if deprecated_in else "",
                "removed_in":
                    "\n   This will be removed in %s." %
                    removed_in if removed_in else "",
                "details":
                    " %s" % details if details else ""}

            deprecation_note = (".. deprecated::{deprecated_in}"
                                "{removed_in}{details}".format(**parts))

            # default location for insertion of deprecation note
            loc = 1

            # split docstring at first occurrence of newline
            string_list = existing_docstring.split("\n", 1)

            if len(string_list) > 1:
                # With a multi-line docstring, when we modify
                # existing_docstring to add our deprecation_note,
                # if we're not careful we'll interfere with the
                # indentation levels of the contents below the
                # first line, or as PEP 257 calls it, the summary
                # line. Since the summary line can start on the
                # same line as the """, dedenting the whole thing
                # won't help. Split the summary and contents up,
                # dedent the contents independently, then join
                # summary, dedent'ed contents, and our
                # deprecation_note.

                # in-place dedent docstring content
                string_list[1] = textwrap.dedent(string_list[1])

                # we need another newline
                string_list.insert(loc, "\n")

                # change the message_location if we add to end of docstring
                # do this always if not "top"
                if message_location != "top":
                    loc = 3

            # insert deprecation note and dual newline
            string_list.insert(loc, deprecation_note)
            string_list.insert(loc, "\n\n")

            function.__doc__ = "".join(string_list)

        @functools.wraps(function)
        def _inner(*args, **kwargs):
            if should_warn:
                if is_unsupported:
                    cls = UnsupportedWarning
                else:
                    cls = DeprecatedWarning
                warnings.simplefilter('always', DeprecationWarning)
                the_warning = cls(function.__name__, deprecated_in,
                                  removed_in, details)
                warnings.warn(the_warning, category=DeprecationWarning,
                              stacklevel=3)
                warnings.resetwarnings()
            return function(*args, **kwargs)
        return _inner
    return _function_wrapper


"""
The arcgis.geoanalytics module provides types and functions for distributed analysis of large datasets.
These GeoAnalytics tools work with big data registered in the GISs datastores as well as with feature layers.

Use arcgis.geoanalytics.is_analysis_supported(gis) to check if geoanalytics is supported in your GIS.

Note: GeoAnalytics operations use the following context parameters defined in the `arcgis.env` module:

        =========================     ====================================================================
        **Context parameter**         **Description**
        -------------------------     --------------------------------------------------------------------
        out_spatial_reference         Used for setting the output spatial reference
        -------------------------     --------------------------------------------------------------------
        process_spatial_reference     Used for setting the processing spatial reference.
        -------------------------     --------------------------------------------------------------------
        analysis_extent               Used for setting the analysis extent.
        -------------------------     --------------------------------------------------------------------
        output_datastore              Used for setting the output datastore to be used.
        =========================     ====================================================================
"""

from . import summarize_data, analyze_patterns, use_proximity, manage_data, find_locations, data_enrichment

def get_datastores(gis=None):
    """
    Returns a helper object to manage geoanalytics datastores in the GIS.
    If a gis isn't specified, returns datastore manager of arcgis.env.active_gis
    """
    import arcgis
    gis = arcgis.env.active_gis if gis is None else gis

    for ds in gis._datastores:
        if 'GeoAnalytics' in ds._server['serverFunction']:
            return ds

    return None

def is_supported(gis=None):
    """
    Returns True if the GIS supports geoanalytics. If a gis isn't specified,
    checks if arcgis.env.active_gis supports geoanalytics
    """
    import arcgis
    gis = arcgis.env.active_gis if gis is None else gis
    if 'geoanalytics' in gis.properties.helperServices:
        return True
    else:
        return False

